import inspect
from collections import OrderedDict
from peepshow.utils.python import id_to_str, hash_to_str

def is_bound(obj):
    try:
        return hasattr(obj, '__self__')
    except Exception:
        return False

def is_subscribable(obj):
    try:
        return is_bound(obj.__getitem__)
    except Exception:
        return False

def is_iterable(obj):
    try:
        return is_bound(obj.__iter__)
    except Exception:
        return False

is_callable = lambda obj: callable(obj)

builtin_types = (
type,
type(Ellipsis),
type(None),
bool,
str,
int, float, complex,
list, tuple, set, dict, frozenset,
range, slice,
bytes, bytearray, memoryview)

def is_of_builtin_type(obj):
    for t in builtin_types:
        if isinstance(obj, t):
            return True, t, t.__name__
    return False, None, ''

def can_be_called_wo_args(func):
    if not is_callable(func):
        return False
    try:
        spec = inspect.getfullargspec(func)
    except:
        return False

    len_ = lambda x: 0 if x is None else len(x)

    return len(spec.args) <= len_(spec.defaults) + int(is_bound(func)) and \
           len(spec.kwonlyargs) == len_(spec.kwonlydefaults)

class Feature:
    def __init__(self, value, exists):
        self.value = value
        self.exists = exists

    def __str__(self):
        if self.exists:
            return str(self.value)
        else:
            return 'N/A'

def get_default_action(obj):
    type_is_builtin, type_, _ = is_of_builtin_type(obj)

    if type_ == dict:
        return '**'
    if type_ in (list, tuple, set, dict, frozenset):
        return '*'
    if type_ is type:
        return 'd'
    if type_is_builtin:
        return '!_'
    if is_iterable(obj):
        return '*'
    if can_be_called_wo_args(obj):
        return '()'
    return 'd'


def read_features(obj, feat_list):
    def try_read(feat):
        try:
            value = feat(obj)
            exists = True
        except Exception:
            value = None
            exists = False
        return Feature(value, exists)

    return OrderedDict( (feat[0], try_read(feat[1])) for feat in feat_list )


class Text:
    def __init__(self, text, *, hlight):
        self.text = text
        self.hlight = hlight

    def __str__(self):
        return self.text

    def __bool__(self):
        return self.hlight

def str_vs_repr(obj):
    str_obj = str(obj)
    if obj.__repr__ is obj.__str__:
        return Text('is REPR', hlight=False)
    elif str_obj == repr(obj):
        return Text('REPR', hlight=False)
    else:
        return Text(repr(str_obj), hlight=True)

def qualname_vs_name(obj):
    if obj.__qualname__ == obj.__name__:
        return Text('NAME', hlight=False)
    else:
        return Text(repr(obj.__qualname__), hlight=True)


class HexInt:
    def __init__(self, x):
        self.x = int(x)

    def __str__(self):
        return str(self.x) + ' = ' + hex(self.x)


MAJOR_FEATURES = [
        ('TYPE',          type),
        ('REPR',          lambda obj: repr(repr(obj)) ),
        ('STR',           str_vs_repr),
        ('NAME',          lambda obj: repr(obj.__name__)),
        ('QUALNAME',      qualname_vs_name),
        ('SIGNATURE',     inspect.signature),
        ('CALLABLE',      is_callable),
        ('ITERABLE',      is_iterable),
        ('SUBSCRIBABLE',  is_subscribable),
        ('LEN',           len),
        ('BOOL',          bool),
        ('INT',           HexInt),
        ('ID',            lambda obj: Text(id_to_str(id(obj)), hlight=False)),
        ('HASH',          lambda obj: Text(hash_to_str(hash(obj)), hlight=False)),
        ('BINDING',       lambda obj: obj.__self__),
        ('BASES',         lambda obj: Text(str(obj.__bases__), hlight=False)),
        ('MRO',           lambda obj: Text(str(obj.mro()), hlight=False)),
    ]

MINOR_FEATURES = [
        ('IsModule',            inspect.ismodule),
        ('IsClass',             inspect.isclass),
        ('IsMethod',            inspect.ismethod),
        ('IsFunction',          inspect.isfunction),
        ('IsGeneratorFunction', inspect.isgeneratorfunction),
        ('IsGenerator',         inspect.isgenerator),
        ('IsCoroutineFunction', inspect.iscoroutinefunction),
        ('IsCoroutine',         inspect.iscoroutine),
        ('IsAwaitable',         inspect.isawaitable),
        ('IsAsyncGenFunction',  inspect.isasyncgenfunction),
        ('IsAsyncGen',          inspect.isasyncgen),
        ('IsTraceback',         inspect.istraceback),
        ('IsFrame',             inspect.isframe),
        ('IsCode',              inspect.iscode),
        ('IsBuiltIn',           inspect.isbuiltin),
        ('IsRoutine',           inspect.isroutine),
        ('IsAbstract',          inspect.isabstract),
        ('IsMethodDescriptor',  inspect.ismethoddescriptor),
        ('IsDataDescriptor',    inspect.isdatadescriptor),
        ('IsGetSetDescriptor',  inspect.isgetsetdescriptor),
        ('IsMemberDescriptor',  inspect.ismemberdescriptor),
    ]



import inspect
from functools import partial
from peepshow.pager import cache as paged_cache
from peepshow.pager.cache import PagedCache
from peepshow.core.probes import is_subscribable, is_iterable, is_callable
from peepshow.utils.terminal import style, Style, Fore, Back
from peepshow.core.probes import Text, is_of_builtin_type
from peepshow.core.exceptions import CommandError

def get_signature(obj):
    try:
        return str(inspect.signature(obj))
    except:
        return '(?)'

def str_func(mode, item):
    name, transformation = item
    attr = transformation.result
    available = transformation.available

    base = None
    if available:
        braces = ''
        if is_callable(attr):
            try:
                builtin_name = attr.__name__
            except AttributeError:
                pass
            else:
                if builtin_name != name:
                    braces += builtin_name

            braces += get_signature(attr)
        if is_subscribable(attr):
            braces += '[]'
        if is_iterable(attr):
            braces += '*'

        if name or braces:
            name_color = fr'{style(Style.BRIGHT, name)}' if name else ''
            sep = ':' if mode == 'dict' else ''
            braces_color = fr'{style(Fore.LIGHTGREEN_EX, braces)}' if braces else ''
            base = ' '.join(x for x in (name_color, sep, braces_color) if x)

        if attr is None:
            type_name = None
            value = repr(attr)
        elif is_callable(attr):
            if inspect.isclass(attr):
                try:
                    type_name = type(attr).__name__
                except:
                    type_name = '?'
                type_name = f'<{type_name}>'
                type_name = style(Fore.LIGHTRED_EX, type_name)
                value = repr(str(attr))
            else:
                type_name = None
                value = None
        elif is_of_builtin_type(attr)[0]:
            type_name = None
            value = repr(attr).splitlines()[0] # what if multiline
        else:
            try:
                type_name = type(attr).__name__
            except:
                type_name = '?'
            type_name = f'<{type_name}>'
            type_name = style(Fore.LIGHTRED_EX, type_name)
            value = repr(str(attr))

        if value is not None:
            value = style(Fore.LIGHTBLUE_EX, value)

        info = [base, type_name, value]
    else:
        if name:
            base = fr'{style(Style.BRIGHT, name)}'
        info = [base, 'N/A']

    return ' '.join( (x for x in info if x) )


class Explorer:
    def __init__(self, ctx):
        self.ctx = ctx
        self.cache = PagedCache([])
        self.cached_target = None

    def fill(self, content, style, offset=0):
        """style: 'attr', 'list', 'dict'
        """
        str_func_styled = partial(str_func, style)
        try:
            self.cache = paged_cache.page(content, str_func_styled, offset)
        except paged_cache.TooManyInitialIterations as ex:
            raise CommandError(ex) from ex
        self.cached_target = self.ctx.target

    def recall(self):
        if self.ctx.target is not self.cached_target:
            raise CommandError('No cache for this target')
        self.cache.recall_cache()

    def get_transformation(self, index):
        if self.ctx.target is not self.cached_target:
            raise CommandError('No cache for this target')

        try:
            key, transformation = self.cache[index]
        except Exception as ex:
            raise CommandError(ex) from ex

        if not transformation.available:
            raise CommandError('Item not available')

        return transformation



import sys
import os
from math import ceil, log2

OS_IS_WINDOWS = os.name == 'nt'

# https://docs.python.org/3/library/platform.html#cross-platform
# read paragraph on platform.architecture()
OS_BITS = 2**ceil(log2(log2(sys.maxsize)))

dev_mode_enabled = bool(int(os.getenv('PEEPSHOW_DEV_MODE', 0)))



import inspect
import ast
import astunparse
import astor
import pprintpp
from peepshow.utils.system import OS_BITS
from textwrap import dedent
from peepshow.core.env import Env
from functools import wraps
from itertools import islice
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import TerminalFormatter

def always_assert(condition):
    """Assert which works also when code optimization is enabled"""
    if not condition:
        raise AssertionError


def caller_gloloc(level=2):
    """Return globals & locals of the caller of the caller.

    level: 0 is current frame, 1 is the caller, 2 is caller of the caller
    """
    caller_frame_info = inspect.stack()[level]
    caller_frame = caller_frame_info.frame
    glo = caller_frame.f_globals
    loc = caller_frame.f_locals
    return Env(glo, loc)


def arg_names(level=2):
    """Try to determine names of the variables given as arguments to the caller
    of the caller. This works only for trivial function invocations. Otherwise
    either results may be corrupted or exception will be raised.

    level: 0 is current frame, 1 is the caller, 2 is caller of the caller
    """
    try:
        caller_frame_info = inspect.stack()[level]
        caller_context = caller_frame_info.code_context
        code = dedent(''.join(caller_context))
        tree = ast.parse(code, '', 'eval')
        always_assert(isinstance(tree.body, ast.Call))
        args = tree.body.args
        names = [astunparse.unparse(arg).strip() for arg in args]
        return names
    except Exception as ex:
        raise Exception('Cannot determine arg names') from None


def id_to_str(id_):
    """Return id as fixed-length hex."""
    if id_ < 0:
        id_ += 2 ** OS_BITS
    return f'0x{id_:0{OS_BITS//4}x}'

def hash_to_str(hash_):
    """Return hash as fixed-length hex."""
    return id_to_str(hash_)


class InvocationError(RuntimeError):
    pass

class CheckInvocation:
    """Check if exception comes from the function body or maybe from the fact
    that it is invoked incorrectly.
    """

    def __init__(self, caller_stack_depth=1):
        self.caller_stack_depth = caller_stack_depth

    def __enter__(self):
        pass

    def __exit__(self, ex_type, ex_value, tb):

        def get_stack_depth(tb=None):
            depth = 0
            while True:
                try:
                    tb = tb.tb_next
                except AttributeError:
                    break
                depth += 1
            return depth

        if ex_value:
            exc_stack_depth = get_stack_depth(tb)
            if exc_stack_depth == self.caller_stack_depth:
                raise InvocationError(ex_value) from None

class NoException(RuntimeError):
    def __init__(self, return_value):
        super().__init__("no exception raised")
        self.return_value = return_value

def catch(func):
    """Return wrapper that executes target function and return exception raised
    by this target function. If no exception is raised by the target function,
    wrapper raises NoException."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return_value = func(*args, **kwargs)
        except Exception as ex:
            return ex
        else:
            raise NoException(return_value) from None
    return wrapper

def nth(iterable, index):
    """Get `index`-nth element from iterable."""
    return [*islice(iterable, index, index+1)][0]


def prettify_expr(expr):

    def pretty_string(s, *args, **kwargs):
        return repr(s)

    def pretty_source(source):
        return ''.join(source)

    tree = ast.parse(expr)
    pretty = astor.code_gen.to_source(tree, pretty_string=pretty_string, pretty_source=pretty_source).strip()

    # alternatively do:
    # pretty = black.format_str(expr, line_length=math.inf).strip()

    return pretty

def crayon_expr(expr):
    return highlight(expr, PythonLexer(), TerminalFormatter()).strip()

def exc_to_str(exc, show_type=False):

    def try1():
        return exc.msg

    def try2():
        return str(exc)

    tries = [try1, try2]

    ret = ""
    for t in tries:
        try:
            ret = t()
            if ret:
                break
        except Exception:
            pass

    type_name = type(exc).__name__

    if ret:
        if show_type:
            return f"{type_name}: {ret}"
        else:
            return ret
    else:
        return type_name

def pformat(expr):
    # alternatively: pprint.pformat
    return pprintpp.pformat(expr)



from types import SimpleNamespace

class Env(SimpleNamespace):
    def __init__(self, glo, loc):
        super().__init__(glo=glo, loc=loc)
        self.initial = {**glo, **loc}
        self.current = {**self.initial}

    def update(self, items):
        self.current.update(items)



import ast
import peepshow.core.trans as trans
from peepshow.utils.python import prettify_expr

def _find_var_offsets(obj, var_name):
    ret = ()
    if isinstance(obj, ast.Name):
        if obj.id == var_name:
            ret = ((obj.col_offset,),)
    elif isinstance(obj, ast.AST):
        ret = tuple(_find_var_offsets(getattr(obj, field), var_name) for field in obj._fields)
    elif isinstance(obj, list):
        ret = tuple(_find_var_offsets(item, var_name) for item in obj)
    return sum(ret, ())


def replace_var_name(expr, var_name, replacement):
    tree = ast.parse(expr)
    offsets = _find_var_offsets(tree.body[0], var_name)

    bias = 0
    for start in offsets:
        start_biased = start + bias
        stop_biased = start_biased + len(var_name)
        expr = expr[:start_biased] + replacement + expr[stop_biased:]
        bias += len(replacement) - len(var_name)

    return expr

def is_dependent_expr(expr):
    return replace_var_name(expr, '_', '_') != replace_var_name(expr, '_', '')

def expr_low_prio(expr):
    try:
        kind = ast.parse(expr).body[0].value
    except:
        return True

    high_prio_objects = (
        ast.Num,
        ast.Str,
        ast.Name,
        ast.Dict,
        ast.Set,
        ast.Tuple,
        ast.List,
        ast.Call,
        ast.Subscript,
        ast.Attribute)

    # if anything unexpected comes, priority is assumed low
    return not isinstance(kind, high_prio_objects)

def expr_num_literal(expr):
    try:
        kind = ast.parse(expr).body[0].value
    except:
        return False

    return isinstance(kind, ast.Num)

def add_panth(expr):
    return f"({expr})"

class Dialect:

    def stringify(self, obj):
        return obj.stringify(self)

    def is_prev_low_prio(self, transformation):
        raise NotImplementedError

    def is_prev_num_literal(self, transformation):
        prev_trans = transformation.get_prev()
        if isinstance(prev_trans, (trans.Given, trans.Eval)):
            return expr_num_literal(prev_trans.expr)
        return False

    def apply(self, transformation, prev_str=None):
        cls = type(transformation)
        hndl_name = 'fmt_' + cls.__name__.lower()
        hndl = getattr(self, hndl_name)
        return hndl(x=prev_str, t=transformation)

# Depreciated
class EvaluableCustom(Dialect):

    def is_prev_low_prio(self, transformation):
        prev_trans = transformation.get_prev()
        if prev_trans is None:
            return False
        elif isinstance(prev_trans, (trans.GloLoc,
                                     trans.Iter,
                                     trans.Attrib,
                                     trans.Pass,
                                     trans.PassExpr,
                                     trans.Call,
                                     trans.Subscr,
                                     trans.SubscrExpr,
                                     trans.Iter)):
            return False
        elif isinstance(prev_trans, (trans.Given, trans.Eval)):
            return expr_low_prio(prev_trans.expr)
        else:
            return True

    def add_panth(self, x, t):
        if self.is_prev_low_prio(t) or self.is_prev_num_literal(t):
            return add_panth(x)
        else:
            return x

    def fmt_gloloc(self, x, t):
        return '{**globals(), **locals()}'

    def fmt_given(self, x, t):
        return t.expr

    def fmt_attrib(self, x, t):
        x = self.add_panth(x, t)
        return f'{x}.{t.attr_name}'

    def fmt_pass(self, x, t):
        return f'{t.func_name}({x})'

    def fmt_passexpr(self, x, t):
        return f'{t.func_name}({x})'

    def fmt_call(self, x, t):
        if t.catched:
            return f'catch({x})({t.args_kwargs})'
        else:
            x = self.add_panth(x, t)
            return f'{x}({t.args_kwargs})'

    def fmt_subscr(self, x, t):
        if isinstance(t.get_prev(), trans.GloLoc):
            return f'{t.index}'
        else:
            x = self.add_panth(x, t)
            return f'{x}[{t.index!r}]'

    def fmt_subscrexpr(self, x, t):
        x = self.add_panth(x, t)
        return f'{x}[{t.expr}]'

    def fmt_iter(self, x, t):
        return f'nth({x}, {t.index!r})'

    def fmt_eval(self, x, t):
        if is_dependent_expr(t.expr):
            x = self.add_panth(x, t)
            return replace_var_name(t.expr, '_', x)
        else:
            return t.expr


class EvaluableAuto(Dialect):

    def stringify(self, obj):
        return prettify_expr(obj.stringify(self))

    def fmt_gloloc(self, x, t):
        return '{**globals(), **locals()}'

    def fmt_given(self, x, t):
        return t.expr

    def fmt_attrib(self, x, t):
        x = add_panth(x)
        return f'{x}.{t.attr_name}'

    def fmt_pass(self, x, t):
        return f'{t.func_name}({x})'

    def fmt_passexpr(self, x, t):
        return f'{t.func_name}({x})'

    def fmt_call(self, x, t):
        if t.catched:
            return f'catch({x})({t.args_kwargs})'
        else:
            x = add_panth(x)
            return f'{x}({t.args_kwargs})'

    def fmt_subscr(self, x, t):
        if isinstance(t.get_prev(), trans.GloLoc):
            return f'{t.index}'
        else:
            x = add_panth(x)
            return f'{x}[{t.index!r}]'

    def fmt_subscrexpr(self, x, t):
        x = add_panth(x)
        return f'{x}[{t.expr}]'

    def fmt_iter(self, x, t):
        return f'nth({x}, {t.index!r})'

    def fmt_eval(self, x, t):
        if is_dependent_expr(t.expr):
            x = add_panth(x)
            return replace_var_name(t.expr, '_', x)
        else:
            return t.expr

Evaluable = EvaluableAuto

class Readable(Dialect):

    def is_prev_low_prio(self, transformation):
        prev_trans = transformation.get_prev()
        if prev_trans is None:
            return False
        elif isinstance(prev_trans, (trans.GloLoc,
                                     trans.Iter,
                                     trans.Attrib,
                                     trans.Call,
                                     trans.Subscr,
                                     trans.SubscrExpr)):
            return False
        elif isinstance(prev_trans, (trans.Given, trans.Eval)):
            return expr_low_prio(prev_trans.expr)
        else:
            return True

    def add_panth(self, x, t):
        if self.is_prev_low_prio(t) or self.is_prev_num_literal(t):
            return add_panth(x)
        else:
            return x

    def fmt_gloloc(self, x, t):
        return '<*>'

    def fmt_given(self, x, t):
        return t.expr

    def fmt_attrib(self, x, t):
        x = self.add_panth(x, t)
        return f'{x}.{t.attr_name}'

    def fmt_pass(self, x, t):
        return f'{x} -> {t.func_name}'

    def fmt_passexpr(self, x, t):
        return f'{x} -> {t.func_name}'

    def fmt_call(self, x, t):
        if t.catched:
            return f'catch({x})({t.args_kwargs})'
        else:
            x = self.add_panth(x, t)
            return f'{x}({t.args_kwargs})'

    def fmt_subscr(self, x, t):
        if isinstance(t.get_prev(), trans.GloLoc):
            return f'{t.index}'
        else:
            x = self.add_panth(x, t)
            return f'{x}[{t.index!r}]'

    def fmt_subscrexpr(self, x, t):
        x = self.add_panth(x, t)
        return f'{x}[{t.expr}]'

    def fmt_iter(self, x, t):
        x = self.add_panth(x, t)
        return f'{x}<{t.index}>'

    def fmt_eval(self, x, t):
        expr = prettify_expr(t.expr)
        if is_dependent_expr(t.expr):
            return f'{x} => {expr}'
        else:
            return expr



from getch import getch # py-getch package
import sys
import colorama
import re
from peepshow.utils import terminal
import shutil

class Line:
    ansi_escape = re.compile(r'\x1b[^m]*m')

    def __init__(self, text):
        self.text = text

    def no_colors(self):
        return self.ansi_escape.sub('', self.text)

    def __len__(self):
        return len(self.no_colors())

    def __str__(self):
        return self.text

    def trim(self, length):
        in_buf = self.text
        out_buf = ''
        out_buf_len = 0
        while in_buf and out_buf_len < length:

            m = self.ansi_escape.match(in_buf)
            if m:
                out_buf += m.group()
                in_buf = in_buf[m.end():]
                continue
            else:
                out_buf += in_buf[0]
                out_buf_len += 1
                in_buf = in_buf[1:]

        return Line(out_buf + colorama.Style.RESET_ALL)

    def __add__(self, other):
        return Line(self.text + str(other))

class Pager:

    def __init__(self, start_page_callback=((lambda: None),), numeric=False):
        self.page_width, self.page_height = shutil.get_terminal_size()
        self.start_page_callback = start_page_callback
        self.numeric = numeric

    def trim_line(self, line):
        elip = Line(colorama.Style.DIM + '...' + colorama.Style.RESET_ALL)
        if len(line) > self.page_width:
            line = line.trim(self.page_width - len(elip)) + elip
            line = line.trim(self.page_width)
        return line

    def print_line(self, line):
        line = self.trim_line(line)
        sys.stdout.write(str(line))
        line_len = len(line)
        last_row_len = line_len % self.page_width
        if not line_len or last_row_len:
            # add extra CR/CRLF if line doesn't cover entire width
            # this does matter under Window and is meaningless under Linus
            print()

    def prompt(self):
        hint = f"Press Q/ESC{['', '/NUMBER'][self.numeric]} to stop or any other key to continue..."
        line = str(self.trim_line(Line(hint)))
        terminal.print_help(line, end='')
        sys.stdout.flush()
        try:
            key = getch()
            ESC = '\x1b'
            CTRL_C = '\x03'
            terminating_keys = [ESC, CTRL_C, 'q', 'Q']
            numbers = [str(x) for x in range(10)]
            if self.numeric:
                terminating_keys += numbers
                if key in numbers:
                    terminal.prefill_input(key)
            interrupted = key in terminating_keys
        except KeyboardInterrupt:
            interrupted = True
        sys.stdout.write('\r' + ' '*(len(line)) + '\r')
        sys.stdout.flush()
        return interrupted

    def execute_start_page_callback(self):
        self.start_page_callback[0](*self.start_page_callback[1:])

    def page(self, lines):
        footer_height = 1 # reserve one line for the prompt
        usable_height = (self.page_height - footer_height)
        self.execute_start_page_callback()

        for line_idx, line in enumerate(lines):
            end_page = (line_idx + 1) % usable_height == 0

            self.print_line(Line(line))

            if end_page:
                if self.prompt():
                    break
                else:
                    self.execute_start_page_callback()
                    terminal.clear()

        sys.stdout.flush()

def page(content):
    p = Pager()
    p.page(content)



from IPython.core.magic import register_line_magic

@register_line_magic
def peep(target):
    """Peep the target.

    Usage:
        %peep target.

    Calls peep() from peepshow library.
    """

    import peepshow
    target = target.strip()
    if target:
        return peepshow.peep_(target)
    else:
        return peepshow.peep_()

del peep
del register_line_magic



import sys

supported_versions = ('3.6', '3.7')
python_version = '.'.join(map(str, sys.version_info[:2]))
if python_version not in supported_versions:
    raise RuntimeError('python version ' + python_version + ' is not supported')

from peepshow.peep import peep, peep_
from peepshow.show import show, show_
from peepshow.utils.python import catch
from peepshow.utils.python import nth



from peepshow.core.exceptions import CommandError
from peepshow.utils.python import CheckInvocation, InvocationError
from peepshow.utils.python import prettify_expr, exc_to_str

class Transformation:
    def __init__(self):
        self.result = None
        self._prev = None
        self._next = None
        self.available = False

    def link(self, prev, *, forward=True):
        self._prev = prev
        if forward:
            prev._next = self

    def evaluate(self, prev, ctx):
        raise NotImplementedError

    def assign(self, result):
        self.result = result
        self.available = True

    def execute(self, prev, ctx, *, safe=False):
        try:
            result = self.evaluate(prev, ctx)
        except:
            if not safe:
                raise
        else:
            self.assign(result)

    def get_prev(self):
        if self._prev:
            return self._prev
        else:
            raise IndexError

    def get_next(self):
        if self._next:
            return self._next
        else:
            raise IndexError

    def get_first(self):
        if self._prev:
            return self._prev.get_first()
        else:
            return self

    def stringify(self, style):
        prev_str = self._prev.stringify(style)
        curr_str = style.apply(self, prev_str)
        return curr_str

class Initial(Transformation):
    def __init__(self, target):
        super().__init__()
        self.result = target

class GloLoc(Initial):
    def stringify(self, style):
        return style.apply(self)

class Given(Initial):
    def __init__(self, target, expr):
        super().__init__(target)
        self.expr = prettify_expr(expr)

    def stringify(self, style):
        return style.apply(self, self.expr)

class Attrib(Transformation):
    def __init__(self, attr_name):
        super().__init__()
        self.attr_name = attr_name

    def evaluate(self, prev, env):
        return getattr(prev.result, self.attr_name)

class Pass(Transformation):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.func_name = func.__name__

    def evaluate(self, prev, ctx):
        return self.func(prev.result)

class PassExpr(Transformation):
    def __init__(self, func_name):
        super().__init__()
        self.func_name = func_name

    def evaluate(self, prev, ctx):
        try:
            value = ctx.eval_(f"{self.func_name}(_)")
        except Exception as ex:
            raise CommandError(f"Invalid invocation: {exc_to_str(ex)}") from ex
        return value

class Call(Transformation):
    def __init__(self, args_kwargs):
        super().__init__()
        self.args_kwargs = args_kwargs
        self.catched = False

    def evaluate(self, prev, ctx):
        call = f'_({self.args_kwargs})'
        try:
            with CheckInvocation(caller_stack_depth=3):
                return ctx.eval_(call)
        except InvocationError as ex:
            raise CommandError(ex) from ex
        except Exception as ex:
            self.catched = True
            return ex

class Subscr(Transformation):
    def __init__(self, index):
        super().__init__()
        self.index = index

    def evaluate(self, prev, ctx):
        return prev.result[self.index]

class SubscrExpr(Transformation):
    def __init__(self, expr):
        super().__init__()
        self.expr = expr

    def evaluate(self, prev, ctx):
        try:
            value = ctx.eval_(f"_[{self.expr}]")
        except Exception as ex:
            raise CommandError(exc_to_str(ex, show_type=True)) from ex
        return value

class Iter(Transformation):
    def __init__(self, index):
        super().__init__()
        self.index = index

    def evaluate(self, prev, ctx):
        return iget(prev.result, self.index)

class Eval(Transformation):
    def __init__(self, expr):
        super().__init__()
        self.expr = expr

    def evaluate(self, prev, ctx):
        try:
            value = ctx.eval_(self.expr)
        except SyntaxError as ex:
            raise CommandError(exc_to_str(ex, show_type=True)) from ex
        return value

class TransformationMgr:
    def __init__(self, target, ctx):
        self.selected = None
        self._ctx = ctx
        self.reset(target)

    def reset(self, transformation):
        self.selected = transformation

    def transform(self, transformation):
        transformation.execute(self.selected, self._ctx)
        self.accept(transformation)

    def accept(self, transformation):
        transformation.link(self.selected)
        self.selected = transformation

    def propose_iter(self, offset):
        for index, item in enumerate(self.selected.result):
            transformation = Iter(index)
            transformation.link(self.selected, forward=False)
            transformation.assign(item)
            yield None, transformation

    def propose_attr(self, attr_names, values=None):
        if values is None:
            for attr_name in attr_names:
                transformation = Attrib(attr_name)
                transformation.link(self.selected, forward=False)
                transformation.execute(self.selected, self._ctx, safe=True)
                yield attr_name, transformation
        else:
            for attr_name, value in zip(attr_names, values):
                transformation = Attrib(attr_name)
                transformation.link(self.selected, forward=False)
                transformation.assign(value)
                yield attr_name, transformation

    def propose_subscr(self, indices, values=None):
        if values is None:
            for index in indices:
                transformation = Subscr(index)
                transformation.link(self.selected, forward=False)
                transformation.execute(self.selected, self._ctx, safe=True)
                yield repr(index), transformation
        else:
            for index, value in zip(indices, values):
                transformation = Subscr(index)
                transformation.link(self.selected, forward=False)
                transformation.assign(value)
                yield repr(index), transformation

    def select_next(self):
        self.selected = self.selected.get_next()

    def select_prev(self):
        self.selected = self.selected.get_prev()

    def select_first(self):
        self.selected = self.selected.get_first()

    def get_history(self, style):
        for item in self:
            yield style.stringify(item), item is self.selected

    def __iter__(self):
        item = self.selected.get_first()
        while True:
            yield item
            try:
                item = item.get_next()
            except IndexError:
                break



from peepshow.utils.python import prettify_expr, pformat

def show(names, values):
    """Show names of variables together with corresponding values."""
    for name, value in zip(names, values):
        print(f'{prettify_expr(name)} = {pformat(value)}')




def peep(*args):
    """Examine local data.
    peep()   # examine all variables in the scope (locals cover globals)
    peep(x)  # examine x (name will be determined only if possible)
    """

    from peepshow.core import peep as core
    from peepshow.utils import python as utils
    from peepshow.core.trans import GloLoc, Given

    if len(args) > 1:
        raise TypeError("Too many arguments.")

    env = utils.caller_gloloc()

    if args:
        expr = utils.arg_names()[0]
        target = Given(args[0], expr)
    else:
        target = GloLoc(env.initial)

    last_target = core.peep(target, env)


def peep_(*args):
    """Examine local data.
    peep_()     # examine all variables in the scope (locals cover globals)
    peep_('x')  # examine x (name will be known as it is explicitely given)
    """

    from peepshow.core import peep as core
    from peepshow.utils import python as utils
    from peepshow.core.trans import GloLoc, Given

    if len(args) > 1:
        raise TypeError("Too many arguments.")

    expr = args[0]
    env = utils.caller_gloloc()

    if args:
        if not isinstance(expr, str):
            raise TypeError("Expression must be a string or None.")

        try:
            target = eval(expr, {}, env.initial)
        except:
            raise SyntaxError('Invalid expression.')

        target = Given(target, expr)
    else:
        target = GloLoc(env.initial)

    last_target = core.peep(target, env)



from peepshow.pager.pager import Pager

class TooManyInitialIterations(StopIteration): pass

class PagedCache:

    def __init__(self, content, str_func=str):
        self.iterator = content.__iter__()
        self.index = 0
        self.offset = 0
        self.cache = []
        self.str_func = str_func

    def iterate(self, times):
        for i in range(times):
            next(self.iterator)
            self.offset += 1
            self.index += 1

    def __iter__(self):
        while True:
            try:
                obj = next(self.iterator)
            except StopIteration:
                break
            line = self.str_func(obj)
            self.cache.append(obj)
            yield f'[{self.index:>5}] {line}'
            self.index += 1

    def clear_cache(self):
        self.offset += len(self.cache)
        self.cache[:] = []

    def __getitem__(self, index):
        cache_index =  index - self.offset
        if not cache_index >= 0 or not cache_index < len(self.cache):
            raise IndexError("You can use only indices visible on the screen.")

        return self.cache[cache_index]

    def recall_cache(self):
        def content_gen():
            for index, entry in enumerate(self.cache):
                key, obj = entry
                line = self.str_func(entry)
                index_ = index + self.offset
                yield f'[{index_:>5}] {line}'
        p = Pager(numeric=True)
        p.page(content_gen())

def page(content, str_func=str, offset=0):
    cache = PagedCache(content, str_func)
    try:
        cache.iterate(offset)
    except StopIteration as ex:
        raise TooManyInitialIterations(f'Only {cache.index} iterations possible.') from ex
    pager = Pager( (PagedCache.clear_cache, cache), numeric=True )
    pager.page(cache)
    return cache



import peepshow.core.dialect as dialect
from peepshow.utils import terminal
from peepshow.core.explorer import Explorer
from peepshow.utils.python import catch, nth
from peepshow.core.trans import TransformationMgr

class Context:
    def __init__(self, target, env):
        self.mgr = TransformationMgr(target, self)
        self._env = env
        self.explorer = Explorer(self)

    @property
    def env(self):
        predefined = {'_': self.target, 'catch': catch, 'nth': nth}
        self._env.update(predefined)
        return self._env

    @property
    def target(self):
        return self.mgr.selected.result

    @property
    def readable(self):
        return dialect.Readable().stringify(self.mgr.selected)

    def eval_(self, expr):
        return eval(expr, {}, self.env.current)

    def exec_(self, expr):
        exec(expr, {}, self.env.current)



import shlex
import sys
from peepshow.cmds.cmds import Commands, CommandError
from peepshow.utils import terminal
from peepshow.utils.terminal import print_error, print_help, style, Fore
from peepshow.utils.python import CheckInvocation, InvocationError
from peepshow.utils.system import dev_mode_enabled
from peepshow.core import goto
from peepshow.core.context import Context
from peepshow.core.probes import get_default_action, is_of_builtin_type
from peepshow.utils.python import exc_to_str

def read_command(ctx):
    try:
        _, _, type_name = is_of_builtin_type(ctx.target)
        if ctx.only_info_visible:
            action = get_default_action(ctx.target)
        else:
            action = 'i'

        type_name_color = style(Fore.LIGHTRED_EX, type_name, for_readline=True)
        type_name_color += ' ' if type_name else ''
        action_color = style(Fore.CYAN, action, for_readline=True)
        action_color += ' ' if action else ''

        cmd_raw = input(type_name_color + action_color + '> ').strip()
        terminal.prefill_input()
        if not cmd_raw:
            # execute suggested action:
            cmd_raw = action
            # just do nothing and print next prompt:
            #raise goto.NextCommand
        try:
            user_inp = shlex.split(cmd_raw, posix=True)
        except Exception as ex:
            print_error(f"Invalid syntax. {exc_to_str(ex)}.")
            raise goto.NextCommand

    except EOFError: # CTRL+D
        raise goto.Stop(exit=False)
    except KeyboardInterrupt: # CTRL+C
        raise goto.Stop(exit=True)

    cmd_alias, *cmd_usr_args = user_inp
    return cmd_raw, cmd_alias, tuple(cmd_usr_args)


def prepare_command(cmds, cmd_raw, cmd_alias, cmd_usr_args):
    try:
        try:
            cmd_hdlr, qualification = cmds[cmd_raw]
            cmd_usr_args = ()
        except KeyError:
            cmd_hdlr, qualification = cmds[cmd_alias]

        if qualification is not None:
            cmd_usr_args = (qualification, *cmd_usr_args)
    except KeyError:
        print_error(f"Unknown command '{cmd_alias}'.")
        cmd_hdlr = Commands.cmd_help
        cmd_usr_args = ()

    cmd_args = (cmds, *cmd_usr_args)
    return cmd_hdlr, cmd_args


def execute_command(ctx, cmd_hdlr, cmd_args):
    while True:
        try:
            ctx.only_info_visible = False

            try:
                with CheckInvocation():
                    new_cmd = cmd_hdlr(*cmd_args)
            except goto.GoToLabel:
                raise
            except InvocationError as ex:
                raise CommandError('Wrong number of arguments.') from ex
            except Exception as ex:
                raise CommandError(ex) from ex

            if new_cmd:
                cmd_hdlr, cmd_args = new_cmd
            else:
                break
        except goto.Stop:
            raise
        except CommandError as ex:
            if dev_mode_enabled:
                raise
            print_error("Error while executing command '{}'.".format(cmd_hdlr.cmd_descr.name))
            print_error(str(ex))
            cmds = cmd_args[0]
            cmd_args = (cmds, cmd_hdlr.cmd_descr.name)
            cmd_hdlr = Commands.cmd_help


def clean_up():
    terminal.clear()


def stop(try_exit):
    def test_exit():
        try:
            exit_func = exit
            return True
        except NameError:
            # e.g. in ipython exit() is not available
            return False

    cancel_stop = False
    if try_exit:
        if sys.gettrace():
            print_error('Underlying debugger cannot be terminated.')
            # in fact it can be, but in case of pdb/ipdb it results in an ugly exception
            print_help("You can use 'Continue' command to close peepshow.")
            cancel_stop = True
        elif test_exit():
            clean_up()
            exit(0)
        else:
            print_error('Underlying Python interpreter cannot be terminated.')
            print_help("You can use 'Continue' command to close peepshow.")
            cancel_stop = True
    else:
        clean_up()

    return cancel_stop


def peep(target, env):
    ctx = Context(target, env)
    terminal.init(ctx.env.current)
    cmds = Commands(ctx)
    execute_command(ctx, Commands.cmd_info, (cmds,))
    while True:
        try:
            cmd_raw, cmd_alias, cmd_usr_args = read_command(ctx)
            cmd_hdlr, cmd_args = prepare_command(cmds, cmd_raw, cmd_alias, cmd_usr_args)
            execute_command(ctx, cmd_hdlr, cmd_args)
        except goto.NextCommand:
            continue
        except goto.Stop as ex:
            if not stop(ex.exit):
                break

    return ctx.target



from setuptools import setup, find_packages
from setuptools.command.install import install
from pathlib import Path
import site

customize_code = """
###############################################################################
# Add peepshow to built-ins
###############################################################################

try:
    import peepshow
    import builtins
    builtins.peep = peepshow.peep
    builtins.show = peepshow.show
    builtins.peep_ = peepshow.peep_
    builtins.show_ = peepshow.show_
except:
    print("peepshow seems to be uninstalled, please manually remove " \\
          "corresponding entry form " + __file__)

###############################################################################
"""

class CustomInstallCommand(install):

    user_options = install.user_options + [
        ('add-builtins', None, 'Add "peep" and "show" keywords to builtins'),
    ]

    def initialize_options(self):
        install.initialize_options(self)
        self.add_builtins = 0

    def do_add_builtins(self):
        if self.user:
            customize_path = Path(site.getusersitepackages()) / 'usercustomize.py'
        else:
            customize_path = Path(site.getsitepackages()[0]) / 'sitecustomize.py'

        with open(customize_path, 'a') as fh:
            fh.write(customize_code)

        print('PeepShow utils added to builtins through: ', customize_path)

    def run(self):
        super().run()
        if self.add_builtins:
            self.do_add_builtins()

readme_path = Path(__file__).parent / 'README.rst'

with open(readme_path) as fh:
    long_description = fh.read()

setup(
    name = 'peepshow',
    version = '0.1.5',
    url = 'https://github.com/gergelyk/peepshow',
    author = 'Grzegorz Krason',
    author_email = 'grzegorz@krason.me',
    description = 'Data Explorer',
    packages = find_packages(),
    keywords = 'debug data explore programming'.split(),
    long_description = long_description,
    python_requires = '>=3.6,<3.8',
    package_data = {'': ['peepshow.1']},
    cmdclass = {
        'install': CustomInstallCommand,
    },
    classifiers = [
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Utilities',
    ],
    install_requires = [
        'astor~=0.7',
        'astunparse~=1.6',
        'colorama~=0.4',
        'getch~=1.0',
        'miscutils~=1.1',
        'pprintpp~=0.4',
        'pygments~=2.2',
        ],
)



import os
from colorama import Fore, Back, Style
import colorama
import rlcompleter
import readline
from peepshow.utils.system import OS_IS_WINDOWS

_completer_suggestions = {}

class Completer(rlcompleter.Completer):
    def complete_ex(self, *args, **kwargs):
        buf = readline.get_line_buffer().strip()
        if buf and buf[0] in ['!', '$']:
            return self.complete(*args, **kwargs)

def update_suggestions(suggestions):
    _completer_suggestions.update(suggestions)

def init(suggestions):
    update_suggestions(suggestions)
    completer = Completer(_completer_suggestions)
    readline.set_completer(completer.complete_ex)
    readline.parse_and_bind("tab: complete")
    colorama.init()
    readline.set_history_length(1000)
    clear()

def clear():
    os.system('cls' if OS_IS_WINDOWS else 'clear')

def print_error(msg='', *args, **kwargs):
    print(style(Back.RED + Fore.WHITE, msg), *args, **kwargs)

def print_help(msg='', *args, **kwargs):
    print(style(Fore.LIGHTYELLOW_EX, msg), *args, **kwargs)

def prefill_input(text=None):
    if text:
        readline.set_startup_hook(lambda: readline.insert_text(text))
    else:
        readline.set_startup_hook()

def style(spec, text, for_readline=False):
    # Thanks to Samuele Santi fot the article:
    # 9https://wiki.hackzine.org/development/misc/readline-color-prompt.html

    RL_PROMPT_START_IGNORE  = '\001'
    RL_PROMPT_END_IGNORE    = '\002'
    term = Style.RESET_ALL

    def wrap(text):
        if for_readline:
            return RL_PROMPT_START_IGNORE + text + RL_PROMPT_END_IGNORE
        else:
            return text

    return wrap(spec) + text + wrap(term)




def show(*args, **kwargs):
    """show(x, y, z=z) # print names & values of arguments
    names of args will be determined only if possible
    names of kwargs will be known as it is explicitely given
    """

    import miscutils.insp as insp
    from peepshow.core import show as core
    from peepshow.utils import python as utils

    if args or kwargs:
        # show specified variables
        if args:
            names = utils.arg_names()
        else:
            names = []
        names += [*kwargs.keys()]
        values = [*args] + [*kwargs.values()]
    else:
        # show all the user variables in scope of the caller
        env = utils.caller_gloloc()
        is_user_var = lambda item: not insp.isaccess(item[0]).special
        user_vars = filter(is_user_var, env.initial.items())
        names, values = zip(*user_vars)

    core.show(names, values)

def show_(*args, **kwargs):
    """show_('x', 'y', z='z') # print names & values of arguments
    values of args and kwargs are expressions to be evaluated in context of the caller
    values of args become names to be displayed
    names of kwargs become names to be displayed
    """

    import miscutils.insp as insp
    from peepshow.core import show as core
    from peepshow.utils import python as utils

    env = utils.caller_gloloc()

    if args or kwargs:
        # show specified variables
        names = [*args] + [*kwargs.keys()]
        exprs = [*args] + [*kwargs.values()]

        for expr in exprs:
            if not isinstance(expr, str):
                raise TypeError("Each expression must be a string.")

        values = [eval(expr, env.glo, env.loc) for expr in exprs]
    else:
        # show all the user variables in scope of the caller
        is_user_var = lambda item: not insp.isaccess(item[0]).special
        user_vars = filter(is_user_var, env.initial.items())
        names, values = zip(*user_vars)

    core.show(names, values)



import re
import os
import shutil
import inspect
from pathlib import Path
from miscutils.insp import isaccess
import peepshow
import peepshow.core.dialect as dialect
import peepshow.core.trans as trans
from peepshow.cmds.base import CommandsBase, command
from peepshow.core import goto, probes
from peepshow.core.probes import read_features, MAJOR_FEATURES, MINOR_FEATURES
from peepshow.core.probes import is_subscribable
from peepshow.core.probes import Text, is_of_builtin_type
from peepshow.utils.terminal import print_help, print_error
from peepshow.utils.terminal import style, Style, Fore, Back
from peepshow.utils import terminal
from peepshow.pager import pager
from peepshow.core.exceptions import CommandError
from peepshow.utils.python import exc_to_str, crayon_expr, pformat

def table_width(table):
    return max(map(len, table)) + 1

def print_target_features(obj, features, *, hlight=False, header=None):
    table = read_features(obj, features)
    width = table_width(table)
    def get_lines():
        if header is not None:
            yield header
        for name, feat in table.items():
            line = f'{name:<{width}}: {feat}'
            if hlight and feat.value:
                line = style(Style.BRIGHT, line)
            yield line
    pager.page(get_lines())

class Commands(CommandsBase):

    def __init__(self, context):
        self.ctx = context

    def qualifier_cmd_eval(alias):
        regexp = re.compile('!(.*)$')
        m = regexp.match(alias)
        return m.groups()[0]

    @command('!<EXPR>', qualifier_cmd_eval)
    def cmd_eval(self, expr):
        """Evaluate <EXPR> and print the result.
        Empty <EXPR> will paste expression of current target and let you edit it.
        See also '$<EXPR>' command.
        """

        expr = expr.strip()
        if expr:
            try:
                try:
                    result = self.ctx.eval_(expr)
                    print(repr(result))
                except SyntaxError:
                    self.ctx.exec_(expr)
            except Exception as ex:
                raise CommandError(exc_to_str(ex, show_type=True)) from ex
            terminal.update_suggestions(self.ctx.env.current)
        else:
            expr = dialect.Evaluable().stringify(self.ctx.mgr.selected)
            terminal.prefill_input('!' + expr)

    def qualifier_cmd_eval_get(alias):
        regexp = re.compile(r'\$(.*)$')
        m = regexp.match(alias)
        return m.groups()[0]

    @command('$<EXPR>', qualifier_cmd_eval_get)
    def cmd_eval_get(self, expr):
        """Evaluate <EXPR> and consider result as a new target.
        Empty <EXPR> will paste expression of current target and let you edit it.
        See also '!<EXPR>' command.
        """

        expr = expr.strip()

        if expr:
            self.ctx.mgr.transform(trans.Eval(expr))
            return Commands.cmd_info, (self,)
        else:
            expr = dialect.Evaluable().stringify(self.ctx.mgr.selected)
            terminal.prefill_input('$' + expr)

    def qualifier_cmd_index(alias):
        regexp = re.compile('\[(.+)\]$')
        m = regexp.match(alias)
        return m.groups()[0]

    @command('[<EXPR>]', qualifier_cmd_index)
    def cmd_index(self, expr):
        """Subscribe target with <EXPR>."""

        if not probes.is_subscribable(self.ctx.target):
            raise CommandError("Item not subscribable.")

        self.ctx.mgr.transform(trans.SubscrExpr(expr))
        return Commands.cmd_info, (self,)

    def qualifier_cmd_call(alias):
        regexp = re.compile('\((.*)\)$')
        m = regexp.match(alias)
        return m.groups()[0]

    @command('(<ARGS>)', qualifier_cmd_call)
    def cmd_call(self, args_kwargs_str):
        """Call target with <ARGS>."""

        if not probes.is_callable(self.ctx.target):
            raise CommandError("Item not callable.")

        try:
            self.ctx.mgr.transform(trans.Call(args_kwargs_str))
        except Exception as ex:
            raise CommandError(f'Invalid invocation: {exc_to_str(ex)}') from ex

        return Commands.cmd_info, (self,)

    @command('ShowCache')
    def cmd_buffer(self):
        """Show content of cache which can be accessed by <INT> command.
        See also '<INT>' command.
        """
        self.ctx.explorer.recall()

    @command('DirAll')
    def cmd_dir_all(self):
        """Call dir(<TARGET>) and show all the results.
        See also 'Dir' and 'DirPublic' commands.
        """
        target = self.ctx.target
        content = self.ctx.mgr.propose_attr(dir(target))
        self.ctx.explorer.fill(content, 'attr')

    @command('Dir')
    def cmd_dir(self):
        """Call dir(<TARGET>) and show results other than of __XYZ__ format.
        See also 'DirAll' and 'DirPublic' commands.
        """
        target = self.ctx.target
        cond = lambda k: not isaccess(k).special
        attr_names = filter(cond, dir(target))
        content = self.ctx.mgr.propose_attr(attr_names)
        self.ctx.explorer.fill(content, 'attr')

    @command('DirPublic')
    def cmd_dir_public(self):
        """Call dir(<TARGET>) and show results other than starting with '_'.
        See also 'Dir' and 'DirAll' commands.
        """
        target = self.ctx.target
        cond = lambda k: isaccess(k).public
        attr_names = filter(cond, dir(target))
        content = self.ctx.mgr.propose_attr(attr_names)
        self.ctx.explorer.fill(content, 'attr')

    @command('Vars')
    def cmd_vars(self):
        """Call vars(<TARGET>) and show all the results.
        See also 'VarsPublic' command.
        """
        target = self.ctx.target
        try:
            v = vars(target)
        except Exception as ex:
            raise CommandError(exc_to_str(ex))

        content = self.ctx.mgr.propose_attr(v.keys(), v.values())
        self.ctx.explorer.fill(content, 'attr')

    @command('VarsPublic')
    def cmd_vars_public(self):
        """Call vars(<TARGET>) and show results other than starting with '_'.
        See also 'Vars' command.
        """
        target = self.ctx.target
        try:
            v = vars(target)
        except Exception as ex:
            raise CommandError(exc_to_str(ex))

        items = ((k, v) for k,v in vars(target).items() if isaccess(k).public)
        try:
            keys, values = zip(*items)
        except ValueError:
            keys, values = (), ()
        content = self.ctx.mgr.propose_attr(keys, values)
        self.ctx.explorer.fill(content, 'attr')

    @command('<INT>', int)
    def cmd_int(self, x):
        """Select <INT>-th item from cache.
        See also 'ShowCache' command.
        """

        transformation = self.ctx.explorer.get_transformation(x)
        self.ctx.mgr.accept(transformation)
        return Commands.cmd_info, (self,)

    @command('..')
    def cmd_back(self):
        """Make one step backward in history.
        See also '-', '/', '.' commands.
        """
        try:
            self.ctx.mgr.select_prev()
            return Commands.cmd_info, (self,)
        except IndexError:
            raise CommandError(f"This is the first item in history. Cannot go back any more.")

    @command('-')
    def cmd_forward(self):
        """Make one step forward in history.
        See also '..', '/', '.' commands.
        """
        try:
            self.ctx.mgr.select_next()
            return Commands.cmd_info, (self,)
        except IndexError:
            raise CommandError("This is the last item in history. Cannot go forward any more.")

    @command('/')
    def cmd_root(self):
        """Go to the begining in history.
        See also '..', '-', '.' commands.
        """
        self.ctx.mgr.select_first()
        return Commands.cmd_info, (self,)

    @command('.')
    def cmd_history(self):
        """Display history.
        See also '..', '-', '/' commands.
        """
        for entry, current in self.ctx.mgr.get_history(dialect.Readable()):
            if current:
                print(style(Style.BRIGHT, entry))
            else:
                print(entry)

    def qualifier_cmd_iterate(alias):
        regexp = re.compile('\*(\d*)$')
        m = regexp.match(alias)
        return int('0' + m.groups()[0])

    @command('*<OFFSET>', qualifier_cmd_iterate)
    def cmd_iterate(self, offset):
        """Iterate target.
        This can be used for listing values of iterable.
        Optional <OFFSET> can be used for skipping certain number of entires.
        """
        target = self.ctx.target
        try:
            content = self.ctx.mgr.propose_iter(offset) # TODO offset
        except TypeError:
            raise CommandError("Item is not iterable.")
        self.ctx.explorer.fill(content, 'list', offset)

    def qualifier_cmd_items(alias):
        regexp = re.compile('\*\*(\d*)$')
        m = regexp.match(alias)
        return int('0' + m.groups()[0])

    @command('**<OFFSET>', qualifier_cmd_items)
    def cmd_items(self, offset):
        """Call <TARGET>.items() and iterate through results.
        This can be used for listing keys and values of the dictionary.
        Optional <OFFSET> can be used for skipping certain number of entires.
        """
        target = self.ctx.target
        try:
            keys = target.keys()
        except Exception as ex:
            raise CommandError(exc_to_str(ex))

        try:
            content = self.ctx.mgr.propose_subscr(keys)
        except:
            raise CommandError("Error while obtaining items to iterate.")
        self.ctx.explorer.fill(content, 'dict', offset)

    def qualifier_cmd_attrib(alias):
        regexp = re.compile('\.([A-Za-z_][A-Za-z0-9_]*)$')
        m = regexp.match(alias)
        return m.groups()[0]

    @command('.<ATTRIB>', qualifier_cmd_attrib)
    def cmd_attrib(self, attr_name):
        """Return attribute <ATTRIB> of the target.
        """
        try:
            self.ctx.mgr.transform(trans.Attrib(attr_name))
        except:
            raise CommandError(f"Attribute cannot be obtained.")
        return Commands.cmd_info, (self,)

    @command('Continue')
    def cmd_continue(self):
        """Exit peepshow and continue execution of underlying application.
        Alternatively enter EOF (under Linux press: CTRL+D)
        See also 'Quit' and 'Export' commands.
        """
        raise goto.Stop(exit=False)

    @command('Quit')
    def cmd_quit(self):
        """Exit peepshow and terminate underlying Python interpreter.
        Alternatively send SIGINT (under Linux press: CTRL+C)
        See also 'Continue' and 'Export' commands.
        """
        raise goto.Stop(exit=True)

    @command('Export')
    def cmd_export(self):
        """Export target expression to IPython.
        This terminates peepshow and initiates next command in underlying IPython
        with the expression describing current target.
        See also 'Continue' and 'Quit' commands.
        """
        try:
            import IPython
            ip = IPython.get_ipython()
            expr = dialect.Evaluable().stringify(self.ctx.mgr.selected)
            ip.set_next_input(expr)
        except:
            raise CommandError("IPython is not available.")
        raise goto.Stop(exit=False)

    @command('?')
    def cmd_docstring(self):
        """Show docstring of the target."""
        try:
            print(inspect.cleandoc(self.ctx.target.__doc__))
        except Exception as ex:
            raise CommandError('Cannot read docstring.')

    @command('??')
    def cmd_source(self):
        """Show source code of the target."""
        try:
            target = self.ctx.target
            file_name = inspect.getsourcefile(target)
            _, first_line_no = inspect.getsourcelines(target)
            first_line_no = max(first_line_no, 1)
            shell_cmd = f'vim -c "map q :qa<CR>" -c ":set number" -RM +{first_line_no} -c "normal zt" {file_name}'
            os.system(shell_cmd)
        except Exception as ex:
            raise CommandError('Cannot find source code.')

    @command('Help', syntax='<CMD>')
    def cmd_help(self, alias=None):
        """Show list of available commands or context help on given command.
        <CMD> is optional and can be either full command name or just a nick
        (capital leters from the name).
        See also 'MANual' command.
        """
        if alias is None:
            print_help('Available commands:')
            print_help()

            def get_summary(cmd):
                try:
                    docstr = inspect.cleandoc(cmd.__doc__)
                except:
                    return ''
                return (docstr + '\n').splitlines()[0]


            def get_header(cmd):
                return ' '.join( (p for p in (cmd.cmd_descr.name, cmd.cmd_descr.syntax) if p))

            entires = [(get_header(cmd), get_summary(cmd)) for cmd in self]
            headers, summaries = zip(*entires)
            left_margin = '  '
            col_sep = '    '
            left_col_width = max(map(len, headers))
            lines = (f"{left_margin}{header:<{left_col_width}}{col_sep}{summary}" for header, summary in entires)
            print_help('\n'.join(sorted(lines, key=lambda x: x.lower())))
            print_help()
            print_help('Use either full command name or just capital leters from it (nick).')
            print_help('Whatever you type, it is case-insensitive.')
            print_help('E.g. you can invoke help by typing: h, H, help, Help, HeLp')
            print_help()
            print_help('Some of the commands have also keyboard bindings assigned.')
            print_help('Invoke "help CMD" for more help on specific command.')
        else:
            try:
                cmd_hdlr, _ = self[alias]
            except KeyError:
                cmds_custom = {cmd.cmd_descr.name.lower(): cmd for cmd in self if callable(cmd.cmd_descr.qualifier)}
                try:
                    cmd_hdlr = cmds_custom[alias.lower()]
                except KeyError:
                    raise CommandError(f'No such command: {alias}.')

            if cmd_hdlr.__doc__:
                parts = ('Syntax:', cmd_hdlr.cmd_descr.name, cmd_hdlr.cmd_descr.syntax)
                print_help(' '.join( (p for p in parts if p) ))
                print_help(inspect.cleandoc(cmd_hdlr.__doc__))
            else:
                raise CommandError(f"No help on '{cmd_hdlr.cmd_descr.name}'.")

    @command('MANual')
    def cmd_man(self):
        """Show manual of peepshow.
        See also 'Help' command.
        """
        man_path = Path(peepshow.__path__[0]) / 'peepshow.1'
        os.system(f'man -l {man_path}')

    @command('Info')
    def cmd_info(self):
        """Show basic information about the target object.
        See also 'Features' command.
        """
        expr = f'{self.ctx.readable:^{shutil.get_terminal_size().columns}}'
        header = style(Back.LIGHTBLUE_EX + Fore.WHITE + Style.BRIGHT, expr)
        terminal.clear()
        print_target_features(self.ctx.target, MAJOR_FEATURES, hlight=True, header=header)
        self.ctx.only_info_visible = True

    @command('Features')
    def cmd_features(self):
        """Show detailed features of the target.
        Results are obtained by 'inspect' module.
        See also 'Info' command.
        """
        print_target_features(self.ctx.target, MINOR_FEATURES, hlight=True)

    @command('CLear')
    def cmd_clear(self):
        """Clear the terminal.
        Alternatively (under Linux) press: CTRL+L
        """
        terminal.clear()

    @command('Type')
    def cmd_type(self):
        """Type of the target becomes a new target."""
        self.ctx.mgr.transform(trans.Pass(type))
        return Commands.cmd_info, (self,)

    @command('Pass', syntax='<FUNC>')
    def cmd_pass(self, func_name):
        """Pass target as an argument to a function."""
        self.ctx.mgr.transform(trans.PassExpr(func_name))
        return Commands.cmd_info, (self,)

    @command('Self')
    def cmd_self(self):
        """Bound object becomes a new target."""
        try:
            self.ctx.mgr.transform(trans.Attrib('__self__'))
        except:
            raise CommandError(f"This target doesn't have bound object.")
        return Commands.cmd_info, (self,)

    @command('Bases')
    def cmd_bases(self):
        """Tuple of base classes of the target becomes a new target.
        See also 'Mro' command.
        """
        try:
            self.ctx.mgr.transform(trans.Attrib('__bases__'))
        except:
            raise CommandError(f"This target doesn't have bases.")
        return Commands.cmd_info, (self,)

    @command('Mro')
    def cmd_mro(self):
        """MRO list of the target becomes a new target.
        See also 'Bases' command.
        """
        try:
            self.ctx.mgr.transform(trans.Attrib('__mro__'))
        except:
            raise CommandError(f'There is no MRO list corresponding to this target.')
        return Commands.cmd_info, (self,)

    @command('eXpression')
    def cmd_expression(self):
        """Print expression that evaluates to examined target.
        See also 'PrettyPrint' command.
        """
        expr = dialect.Evaluable().stringify(self.ctx.mgr.selected)
        print(crayon_expr(expr))

    @command('PrettyPrint')
    def cmd_pretty_print(self):
        """Print target recursively.
        See also 'eXpression' command.
        """
        print(pformat(self.ctx.target))




# command name: e.g. DoSomething
# command nick: e.g. ds
# command alias: e.g. dosomething, DOSOMETHING, ds, DS, ...

from types import SimpleNamespace

def get_nick(name):
    return ''.join(filter(lambda k: 'A' <= k <= 'Z', name)).lower()

def command(name, qualifier=None, syntax=None):
    cmd_descr = SimpleNamespace(**locals())
    cmd_descr.nick = get_nick(name)

    def decorator(func):
        func.cmd_descr = cmd_descr
        return func
    return decorator


class CommandsBase:

    def __iter__(self):
        attrs = vars(type(self)).values()
        attrs = filter(callable, attrs)
        attrs = filter(lambda attr: hasattr(attr, 'cmd_descr'), attrs)
        yield from attrs

    def __getitem__(self, alias):
        if not alias:
            raise KeyError

        cmds_simple = [cmd for cmd in self if cmd.cmd_descr.qualifier is None]
        qualification = None

        # try to find command by name
        cmds_simple_by_name = {cmd.cmd_descr.name.lower(): cmd for cmd in cmds_simple}
        try:
            return cmds_simple_by_name[alias.lower()], qualification
        except KeyError:
            pass

        # try to find command by alias
        cmds_simple_by_nick = {cmd.cmd_descr.nick: cmd for cmd in cmds_simple}
        try:
            return cmds_simple_by_nick[alias.lower()], qualification
        except KeyError:
            pass

        # try to find command by qualifier
        cmds_custom = (cmd for cmd in self if callable(cmd.cmd_descr.qualifier))
        for cmd in cmds_custom:
            try:
                qualification = cmd.cmd_descr.qualifier(alias)
                return cmd, qualification
            except:
                pass

        raise KeyError
